@@ -7,7 +7,7 @@
  *                            | (__| |_| |  _ <| |___
  *                             \___|\___/|_| \_\_____|
  *
- * Copyright (C) 1998 - 2022, Daniel Stenberg, <daniel@haxx.se>, et al.
+ * Copyright (C) Daniel Stenberg, <daniel@haxx.se>, et al.
  *
  * This software is licensed as described in the file COPYING, which
  * you should have received as part of this distribution. The terms
@@ -20,6 +20,8 @@
  * This software is distributed on an "AS IS" basis, WITHOUT WARRANTY OF ANY
  * KIND, either express or implied.
  *
+ * SPDX-License-Identifier: curl
+ *
  ***************************************************************************/
 
 
@@ -27,6 +29,7 @@
 /*       lib/config-dos.h - Hand crafted config file for DOS        */
 /* ================================================================ */
 
+#ifndef OS
 #if defined(DJGPP)
   #define OS  "MSDOS/djgpp"
 #elif defined(__HIGHC__)
@@ -34,35 +37,30 @@
 #else
   #define OS  "MSDOS/?"
 #endif
+#endif
 
 #define PACKAGE  "curl"
 
+#define USE_MANUAL 1
+
 #define HAVE_ARPA_INET_H       1
-#define HAVE_ASSERT_H          1
-#define HAVE_ERRNO_H           1
 #define HAVE_FCNTL_H           1
 #define HAVE_FREEADDRINFO      1
 #define HAVE_GETADDRINFO       1
-#define HAVE_GETPROTOBYNAME    1
 #define HAVE_GETTIMEOFDAY      1
 #define HAVE_IO_H              1
-#define HAVE_IOCTL             1
 #define HAVE_IOCTL_FIONBIO     1
 #define HAVE_IOCTLSOCKET       1
 #define HAVE_IOCTLSOCKET_FIONBIO   1
 #define HAVE_LOCALE_H          1
 #define HAVE_LONGLONG          1
-#define HAVE_MEMORY_H          1
 #define HAVE_NETDB_H           1
 #define HAVE_NETINET_IN_H      1
 #define HAVE_NETINET_TCP_H     1
 #define HAVE_NET_IF_H          1
-#define HAVE_PROCESS_H         1
 #define HAVE_RECV              1
-#define HAVE_RECVFROM          1
 #define HAVE_SELECT            1
 #define HAVE_SEND              1
-#define HAVE_SETJMP_H          1
 #define HAVE_SETLOCALE         1
 #define HAVE_SETMODE           1
 #define HAVE_SIGNAL            1
@@ -71,26 +69,21 @@
 #define HAVE_STRICMP           1
 #define HAVE_STRTOLL           1
 #define HAVE_STRUCT_TIMEVAL    1
-#define HAVE_STRUCT_IN6_ADDR   1
 #define HAVE_SYS_IOCTL_H       1
 #define HAVE_SYS_SOCKET_H      1
 #define HAVE_SYS_STAT_H        1
 #define HAVE_SYS_TYPES_H       1
-#define HAVE_TIME_H            1
 #define HAVE_UNISTD_H          1
 
 #define NEED_MALLOC_H          1
 
 #define SIZEOF_INT             4
 #define SIZEOF_LONG            4
-#define SIZEOF_LONG_DOUBLE     16
-#define SIZEOF_SHORT           2
 #define SIZEOF_SIZE_T          4
-#define SIZEOF_CURL_OFF_T      4
+#define SIZEOF_CURL_OFF_T      8
 #define STDC_HEADERS           1
-#define TIME_WITH_SYS_TIME     1
 
-/* Qualifiers for send(), recv(), and recvfrom() */
+/* Qualifiers for send() and recv() */
 
 #define SEND_TYPE_ARG1         int
 #define SEND_QUAL_ARG2         const
@@ -105,32 +98,11 @@
 #define RECV_TYPE_ARG4         int
 #define RECV_TYPE_RETV         int
 
-#define RECVFROM_TYPE_ARG1     int
-#define RECVFROM_TYPE_ARG2     void
-#define RECVFROM_TYPE_ARG3     int
-#define RECVFROM_TYPE_ARG4     int
-#define RECVFROM_TYPE_ARG5     struct sockaddr
-#define RECVFROM_TYPE_ARG6     int
-#define RECVFROM_TYPE_RETV     int
-#define RECVFROM_TYPE_ARG2_IS_VOID 1
-
 #define BSD
 
 /* CURLDEBUG definition enables memory tracking */
 /* #define CURLDEBUG */
 
-/* USE_ZLIB on cmd-line */
-#ifdef USE_ZLIB
-  #define HAVE_ZLIB_H            1
-  #define HAVE_LIBZ              1
-#endif
-
-/* USE_OPENSSL on cmd-line */
-#ifdef USE_OPENSSL
-  #define HAVE_CRYPTO_CLEANUP_ALL_EX_DATA 1
-  #define OPENSSL_NO_KRB5        1
-#endif
-
 /* to disable LDAP */
 #define CURL_DISABLE_LDAP        1
 
@@ -150,7 +122,6 @@
   #define HAVE_SIGSETJMP  1
   #define HAVE_SYS_TIME_H 1
   #define HAVE_TERMIOS_H  1
-  #define HAVE_VARIADIC_MACROS_GCC 1
 
 #elif defined(__HIGHC__)
   #define HAVE_SYS_TIME_H 1
