@@ -50,7 +50,6 @@ struct per_transfer {
   struct OutStruct outs;
   struct OutStruct heads;
   struct OutStruct etag_save;
-  struct InStruct input;
   struct HdrCbData hdrcbdata;
   long num_headers;
   bool was_last_header_empty;
@@ -68,12 +67,14 @@ struct per_transfer {
   curl_off_t dlnow;
   curl_off_t ultotal;
   curl_off_t ulnow;
+  curl_off_t uploadfilesize; /* expected total amount */
+  curl_off_t uploadedsofar; /* amount delivered from the callback */
   bool dltotal_added; /* if the total has been added from this */
   bool ultotal_added;
 
   /* NULL or malloced */
   char *uploadfile;
-  char *errorbuffer; /* alloced and assigned while this is used for a
+  char *errorbuffer; /* allocated and assigned while this is used for a
                         transfer */
 };
 
